import matplotlib.pyplot as plt
import torch
import tqdm

from utils import kinematics, ldvm, results, mlp, controller, aeroelasticity
import numpy as np


# Kinematics
frequency = 5.878
pitching_amplitude = 79.5
rotation_axis_in_percentage_chord = 0.32
chord = 0.25
u_inf = 10.0

# Dynamical System
m = 1.9
K_h = 2500.0
C_h = 15.877
K_theta = 0.0
C_theta = 0.0
theta_0 = 79.5
xf_c = 0.32

reduced_frequency = ((2 * np.pi * frequency) * (chord / 2)) / u_inf

model_name_list = ['outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_0',
                   'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_1',
                   'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_2',
                   'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_3',
                   'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_4',
                   'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_5',
                   'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_6']

for i in tqdm.tqdm(range(len(model_name_list))):

    model_name = model_name_list[i]
    inputs_variables = ["x_f", "f", "theta_0"]
    outputs_variable = ["lift_coefficient", "moment_coefficient"]

    n_input, n_output, layers_configuration = mlp.load_parameters(neural_networks_path='mlp/files/neural_networks/',
                                                                  model_name=model_name)

    model = mlp.MLP(n_input=n_input,
                    n_output=n_output,
                    layers_configuration=layers_configuration)

    model.load_state_dict(torch.load('mlp/files/neural_networks/' + model_name + '/model_' + model_name + '.pt',
                                     weights_only=True))
    model.eval()

    kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                           nb_time_step_per_period=1000,
                                           input_variable_names=inputs_variables,
                                           output_variable_names=outputs_variable,
                                           variable_values=[rotation_axis_in_percentage_chord, frequency, pitching_amplitude],
                                           m=3,
                                           tau=9,
                                           nb_period=6)
    kinematics_obj.generate_kinematics()

    controller_obj = controller.PIDController(K_p=np.array([0.0, 500.0]),
                                              K_i=np.array([0.0, 75.0]),
                                              K_d=np.array([0.0, 10.0]),
                                              mode='semi_active')

    dynamical_system_obj = aeroelasticity.DynamicalSystem(kinematics_obj=kinematics_obj,
                                                          model=model,
                                                          controller=controller_obj,
                                                          m=m,
                                                          K_h=K_h,
                                                          K_theta=K_theta,
                                                          C_h=C_h,
                                                          C_theta=C_theta)
    try:
        dynamical_system_obj.motion_solver()
    except:
        pass

    kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                           nb_time_step_per_period=1000,
                                           input_variable_names=["x_f", "f", "theta_0"],
                                           output_variable_names=["lift_coefficient", "moment_coefficient"],
                                           variable_values=[rotation_axis_in_percentage_chord, frequency, pitching_amplitude],
                                           heaving_amplitude_chord=0.5,
                                           m=3,
                                           tau=9,
                                           input_time_arr=dynamical_system_obj.time_period,
                                           input_h_arr=dynamical_system_obj.h_period,
                                           input_theta_arr=dynamical_system_obj.theta_period)

    solver_input_data_dict = {
        'model': 'LDVM',
        'movement': {"frequency": kinematics_obj.frequency},
        'name': 'ldvm',
        'solver': {'duration': 2.0 / kinematics_obj.frequency,
                   'nb_time_steps': 500},
        'airfoil': {'aileron_dimension_in_percentage_chord': 0.0,
                    'chord': kinematics_obj.chord,
                    'critical_lesp': 0.14,
                    'rotation_axis_in_percentage_chord': kinematics_obj.rotation_axis_in_percentage_chord},
        'experimental_data': {'drag_coefficient': None,
                              'lift_coefficient': None,
                              'moment_coefficient': None},
        'flow': {'freestream_velocity': kinematics_obj.freestream_velocity,
                 'surge_amplitude': 0.0,
                 'surge_frequency': 0.0}
    }

    results_obj = ldvm.run_case(solver_input_data_dict, kinematics_obj)

    time_ldvm = results_obj.time[results_obj.time >= 1.0 / kinematics_obj.frequency] - 1.0 / kinematics_obj.frequency
    cl_ldvm = results_obj.lift_coefficient[results_obj.time >= 1.0 / kinematics_obj.frequency]
    cm_ldvm = results_obj.moment_airfoil_coefficient[results_obj.time >= 1.0 / kinematics_obj.frequency]

    fig, axs = plt.subplots(1, 2, figsize=(12, 5))

    axs[0].plot(time_ldvm * kinematics_obj.frequency, cl_ldvm, label='LDVM', color='black')
    cl_mlp = dynamical_system_obj.lift_period / (0.5 * 1.225 * (10 ** 2) * 0.25 * 1.0)
    axs[0].plot(dynamical_system_obj.time_period * kinematics_obj.frequency, cl_mlp, label='MLP', color='red')
    axs[0].plot(dynamical_system_obj.time_period * kinematics_obj.frequency, dynamical_system_obj.h_period / chord, linestyle='dashed', color='grey', label='$h/c$')
    axs[0].grid()
    axs[0].legend()
    axs[0].set_xlabel('$\\frac{t}{T}$')
    axs[0].set_ylabel('$C_L$')

    axs[1].plot(time_ldvm * kinematics_obj.frequency, cm_ldvm, label='LDVM', color='black')
    cm_mlp = dynamical_system_obj.moment_period / (0.5 * 1.225 * (10 ** 2) * (0.25 ** 2) * 1.0)
    axs[1].plot(dynamical_system_obj.time_period * kinematics_obj.frequency, cm_mlp, label='MLP', color='red')
    axs[1].grid()
    axs[1].legend()
    axs[1].set_xlabel('$\\frac{t}{T}$')
    axs[1].set_ylabel('$C_M$')

    fig.suptitle('Step ' + str(model_name[-1]) + ' | $U_{\\infty} = $' + str(u_inf) + ' $m.s^{-1}$ | $m = ' + str(m) + '$ $kg.m^{-1}$' + " | $k = $" + str(np.round(reduced_frequency, 3)) + ' | $\\theta_0 = $' + str(np.round(theta_0, 2)) + '$^{\\circ}$ | $x_f/c = $' + str(xf_c) + ' | $C_h = $' + str(np.round(C_h, 3)) + ' $N.s.m^{-2}$ | $K_h = $' + str(np.round(K_h, 3)) + ' $N.m^{-2}$', fontsize=9.5)

    plt.savefig('aeroelasticity/files/vs_ldvm/figures/ldvm_based_dyn.' + model_name + '.png', bbox_inches='tight')
    plt.savefig('aeroelasticity/files/vs_ldvm/figures/ldvm_based_dyn.' + model_name + '.svg', bbox_inches='tight')

plt.show()