import numpy as np
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.termination import get_termination
from pymoo.optimize import minimize
import torch
from utils import mlp, optimization
from pymoo.core.problem import DaskParallelization
from dask.distributed import Client
from dask_mpi import initialize

initialize()

client = Client()
client.restart()

print(client.scheduler_info())

print("DASK STARTED")

runner = DaskParallelization(client)

model_name = 'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_2'
neural_networks_path = 'mlp/files/neural_networks/'

optimization_type = 'dyn'
optimization_name = '3d_test_large'

n_input, n_output, layers_configuration = mlp.load_parameters(neural_networks_path=neural_networks_path,
                                                              model_name=model_name)

model = mlp.MLP(n_input=n_input,
                n_output=n_output,
                layers_configuration=layers_configuration)

model.load_state_dict(torch.load(neural_networks_path + model_name + '/model_' + model_name + '.pt',
                                 weights_only=True))
model.eval()

problem = optimization.DynSingleObjectiveProblem(model,
                                                 xl=np.array([4.0, 5.0, 1000.0]),
                                                 xu=np.array([8.0, 30.0, 4000.0]),
                                                 elementwise_runner=runner)

algorithm = GA(pop_size=960,
               n_offsprings=480,
               sampling=FloatRandomSampling(),
               crossover=SBX(prob=0.9, eta=15),
               mutation=PM(eta=20),
               eliminate_duplicates=True)

termination = get_termination("n_gen", 150)

res = minimize(problem,
               algorithm,
               termination,
               seed=1,
               save_history=True,
               verbose=True)

client.close()
print("DASK SHUTDOWN")

bounds = problem.bounds()

print(res.X)
print(res.F)

optimization.save_results(design_space=res.X,
                          optimization_functions=res.F,
                          bounds=bounds,
                          model_name=model_name,
                          optimization_name=optimization_name,
                          path='optimization/files/' + optimization_type + '/ga/')
