import os

import sklearn

from utils import kinematics, ldvm, mlp, results
import matplotlib.pyplot as plt
import torch
import numpy as np
import os

os.chdir('../')

model_name = 'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_0'
inputs_variables = ["x_f", "f", "theta_0"]
outputs_variable = ["lift_coefficient", "moment_coefficient"]

variables = [0.3, 5.0, 75.0]

kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                       nb_time_step_per_period=1000,
                                       input_variable_names=inputs_variables,
                                       output_variable_names=outputs_variable,
                                       variable_values=variables)

solver_input_data_dict = {
    'model': 'LDVM',
    'movement': {"frequency": kinematics_obj.frequency},
    'name': 'ldvm',
    'solver': {'duration': 2.0 / kinematics_obj.frequency,
               'nb_time_steps': 500},
    'airfoil': {'aileron_dimension_in_percentage_chord': 0.0,
                'chord': kinematics_obj.chord,
                'critical_lesp': 0.18,
                'rotation_axis_in_percentage_chord': kinematics_obj.rotation_axis_in_percentage_chord},
    'experimental_data': {'drag_coefficient': None,
                          'lift_coefficient': None,
                          'moment_coefficient': None},
    'flow': {'freestream_velocity': kinematics_obj.freestream_velocity,
             'surge_amplitude': 0.0,
             'surge_frequency': 0.0}
}

results_obj = ldvm.run_case(solver_input_data_dict, kinematics_obj)

# vertical_extent_airfoil_motion = 2.0 * np.abs(results_obj.h).max()
# power_available_oncoming_flow = 0.5 * 1.225 * 10.0 * 10.0 * 10.0 * vertical_extent_airfoil_motion
# efficiency = np.trapz(results_obj.power_output[250:501], x=results_obj.time[250:501]) / power_available_oncoming_flow

print("LDVM:" + str(results_obj.efficiency))

n_input, n_output, layers_configuration = mlp.load_parameters(neural_networks_path='mlp/files/neural_networks/',
                                                              model_name=model_name)

model = mlp.MLP(n_input=n_input,
                n_output=n_output,
                layers_configuration=layers_configuration)

model.load_state_dict(torch.load('mlp/files/neural_networks/' + model_name + '/model_' + model_name + '.pt',
                                 weights_only=True))
model.eval()

kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                       nb_time_step_per_period=1000,
                                       input_variable_names=inputs_variables,
                                       output_variable_names=outputs_variable,
                                       variable_values=variables,
                                       m=3,
                                       tau=9,
                                       nb_period=2)

mlp_outputs = model(kinematics_obj.get_inputs_mlp())

mlp_outputs = mlp_outputs.detach().numpy()

mlp_results = mlp.MLPResults(kinematics_obj, mlp_outputs)

# mlp_lift_coefficient = np.zeros(501)
# mlp_lift_coefficient[250:501] = mlp_results.heaving_power
# mlp_moment_coefficient = np.zeros(501)
# mlp_moment_coefficient[250:501] = mlp_results.pitching_power

print("MLP:" + str(mlp_results.efficiency))

results_time = results_obj.time[results_obj.time > kinematics_obj.period] - kinematics_obj.period
kinematics_time = kinematics_obj.time[kinematics_obj.time > kinematics_obj.period] - kinematics_obj.period
results_lift_coefficient = results_obj.lift_coefficient[results_obj.time > kinematics_obj.period]
results_moment_airfoil_coefficient = results_obj.moment_airfoil_coefficient[results_obj.time > kinematics_obj.period]
mlp_results_lift_coefficient = mlp_results.lift_coefficient[kinematics_obj.time > kinematics_obj.period]
mlp_results_moment_coefficient = mlp_results.moment_coefficient[kinematics_obj.time > kinematics_obj.period]


x, interp_results_lift_coefficient = results.interp_arr(results_time, results_lift_coefficient, 1000)
r2_score_lift_coefficient = sklearn.metrics.r2_score(interp_results_lift_coefficient, mlp_results_lift_coefficient)
print(r2_score_lift_coefficient)
x, interp_results_moment_airfoil_coefficient = results.interp_arr(results_time, results_moment_airfoil_coefficient, 1000)
r2_score_moment_coefficient = sklearn.metrics.r2_score(interp_results_moment_airfoil_coefficient, mlp_results_moment_coefficient)
print(r2_score_moment_coefficient)

plt.figure()
plt.plot(results_time * kinematics_obj.frequency, results_lift_coefficient, color='green', label='$C_L$ - LDVM')
plt.plot(kinematics_time * kinematics_obj.frequency, mlp_results_lift_coefficient, linestyle='dashed', color='green', label='$C_L$ - MLP - $R2 = $' + str(np.round(r2_score_lift_coefficient, 3)))
plt.grid()
plt.xlabel('$\\frac{t}{T}$')
plt.xlim([0.0, 1.0])
plt.ylabel('$C_L$')
plt.legend(fontsize=9, loc='upper left')

axs_y2 = plt.twinx()
axs_y2.set_ylabel('$C_M$')
axs_y2.plot(results_time * kinematics_obj.frequency, results_moment_airfoil_coefficient, color='red', label='$C_M$ - LDVM')
axs_y2.plot(kinematics_time * kinematics_obj.frequency, mlp_results_moment_coefficient, linestyle='dashed', color='red', label='$C_M$ - MLP - $R2 = $' + str(np.round(r2_score_moment_coefficient, 3)))
axs_y2.legend(fontsize=9, loc='lower right')

reduced_frequency = ((2 * np.pi * kinematics_obj.frequency) * (kinematics_obj.chord / 2)) / kinematics_obj.freestream_velocity

plt.title('$U_{inf} = $' + str(kinematics_obj.freestream_velocity) + ' $m.s^{-1}$ | $k = $' + str(np.round(reduced_frequency, 3)) + ' | $h_0/c = $' + str(np.round(kinematics_obj.heaving_amplitude / kinematics_obj.chord, 2)) + ' | $\\theta_0 = $' + str(np.round(np.degrees(kinematics_obj.pitching_amplitude), 2)) + '$^{\\circ}$ | $x_f/c = $' + str(np.round(kinematics_obj.rotation_axis_in_percentage_chord, 2)), fontsize=9.5)

plt.savefig('mlp/figures/comparison_mlp_ldvm.f_' + str(np.round(variables[1], 2)) + '_xfc_' + str(np.round(variables[0], 2)) + '_theta0_' + str(np.round(variables[2], 2)) + '.svg', bbox_inches='tight')
plt.savefig('mlp/figures/comparison_mlp_ldvm.f_' + str(np.round(variables[1], 2)) + '_xfc_' + str(np.round(variables[0], 2)) + '_theta0_' + str(np.round(variables[2], 2)) + '.png', bbox_inches='tight')

plt.show()