import matplotlib.pyplot as plt
import torch
from utils import mlp, kinematics
import numpy as np
import timeit

model_name = 'outputs_cl_cm_inputs_xf_f_theta_ref_1'
inputs_variables = ["x_f", "f", "theta_0"]
outputs_variable = ["lift_coefficient", "moment_coefficient"]

n_input, n_output, layers_configuration = mlp.load_parameters(neural_networks_path='./files/neural_networks/',
                                                              model_name=model_name)

model = mlp.MLP(n_input=n_input,
                n_output=n_output,
                layers_configuration=layers_configuration)

model.load_state_dict(torch.load('./files/neural_networks/' + model_name + '/model_' + model_name + '.pt',
                                 weights_only=True))
model.eval()

kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                       nb_time_step_per_period=1000,
                                       input_variable_names=inputs_variables,
                                       output_variable_names=outputs_variable,
                                       variable_values=[0.3, 5.0, 70.0],
                                       m=3,
                                       tau=9)

tic = timeit.default_timer()

mlp_outputs = model(kinematics_obj.get_inputs_mlp())

toc = timeit.default_timer()

print("Computational time is: " + str(toc - tic) + ' s.')

mlp_outputs = mlp_outputs.detach().numpy()

mlp_results = mlp.MLPResults(kinematics_obj, mlp_outputs)

fig, axs = plt.subplots(1, 3, figsize=(15, 5))

axs[0].plot(kinematics_obj.time, kinematics_obj.theta, label='$\\theta$')
axs[0].plot(kinematics_obj.time, kinematics_obj.h, label='h')
axs[0].plot(kinematics_obj.time, kinematics_obj.theta_dot, label='$\\dot{\\theta}$')
axs[0].plot(kinematics_obj.time, kinematics_obj.h_dot, label='$\\dot{h}$')
axs[0].set_xlabel('Time (s)')
axs[0].grid()
axs[0].legend()

axs[1].plot(kinematics_obj.time, mlp_results.lift, label='$C_{L}$ | MLP')
axs[1].plot(kinematics_obj.time, mlp_results.moment, label='$C_{M}$ | MLP')
axs[1].set_xlabel('Time (s)')
axs[1].grid()

axs[2].plot(kinematics_obj.time, mlp_results.heaving_power, label='$P_{h}$ | $E_{h}$ = ' + str(np.round(mlp_results.heaving_energy, 4)) + ' J.')
axs[2].plot(kinematics_obj.time, mlp_results.pitching_power, label='$P_{\\theta}$ | $E_{\\theta}$ = ' + str(np.round(mlp_results.pitching_energy, 4)) + ' J.')
axs[2].grid()
axs[2].legend()
axs[2].set_xlabel('Time (s)')
axs[2].set_title('Energy Recoverable: ' + str(np.round(mlp_results.energy_recoverable, 3)) + ' J | Energy Consumed: ' + str(np.round(mlp_results.energy_consumed, 3)) + ' J')

plt.suptitle("$U_{inf} = $" + str(kinematics_obj.freestream_velocity) + " | $f = $" + str(kinematics_obj.frequency) + " | $\\eta = $" + str(np.round(mlp_results.efficiency, 5)), fontsize=10.5)

plt.show()