import matplotlib.pyplot as plt
import numpy as np
import torch
from utils import mlp, kinematics


model_name = 'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_0'
inputs_variables = ["x_f", "f", "theta_0"]
outputs_variable = ["lift_coefficient", "moment_coefficient"]

n_input, n_output, layers_configuration = mlp.load_parameters(neural_networks_path='../mlp/files/neural_networks/',
                                                              model_name=model_name)

model = mlp.MLP(n_input=n_input,
                n_output=n_output,
                layers_configuration=layers_configuration)

model.load_state_dict(torch.load('../mlp/files/neural_networks/' + model_name + '/model_' + model_name + '.pt',
                                 weights_only=True))
model.eval()


theta_0 = np.linspace(0.0, 90.0, 40)
freq_star = np.linspace(0.04, 0.25, 40)

size = (np.size(freq_star), np.size(theta_0))
feathering_parameters = np.zeros(size)
aerodynamic_efficiencies = np.zeros(size)
froude_efficiency = np.zeros(size)
feathering_param_vect = np.zeros(np.size(freq_star))

u_inf = 10.0
h0_c = 0.5
xp_arr = np.linspace(0.1, 0.9, 9)

for xp in xp_arr:
    for i in range(np.size(freq_star)):
        print('Case: ' + str(i+1) + ' / ' + str(np.size(freq_star)))
        for j in range(np.size(theta_0)):
            frequency = (freq_star[i] *  10.0) / 0.25

            kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                                   nb_time_step_per_period=1000,
                                                   input_variable_names=inputs_variables,
                                                   output_variable_names=outputs_variable,
                                                   variable_values=[xp, frequency, theta_0[j]],
                                                   m=3,
                                                   tau=9)

            mlp_outputs = model(kinematics_obj.get_inputs_mlp())

            mlp_outputs = mlp_outputs.detach().numpy()

            mlp_results = mlp.MLPResults(kinematics_obj, mlp_outputs)


            feathering_param_vect[i] = np.degrees(np.arctan((kinematics_obj.heaving_amplitude * kinematics_obj.angular_frequency) / kinematics_obj.freestream_velocity))
            feathering_parameters[i][j] = kinematics_obj.feathering_parameter
            if mlp_results.efficiency >= 0.0:
                aerodynamic_efficiencies[i][j] = mlp_results.efficiency
            else:
                aerodynamic_efficiencies[i][j] = np.nan

    X, Y = np.meshgrid(freq_star, theta_0, indexing='ij')

    X = (X * 10.0) / 0.25
    X = ((2 * np.pi * X) * (kinematics_obj.chord / 2)) / kinematics_obj.freestream_velocity

    red_freq = (freq_star * 10.0) / 0.25
    red_freq = ((2 * np.pi * red_freq) * (kinematics_obj.chord / 2)) / kinematics_obj.freestream_velocity

    plt.figure()
    levels = np.linspace(0.0,0.85,18)
    contour = plt.contourf(X, Y, aerodynamic_efficiencies, levels=levels)
    plt.plot(red_freq, feathering_param_vect, color='red', label='$\\chi = 1$')
    plt.colorbar(contour, label='$\\eta$')
    plt.xlabel('$k$')
    plt.ylabel('$\\theta_{0}$ ($^{\\circ}$)')
    plt.grid()
    plt.title('$U_{\\infty} = $' + str(np.round(u_inf, 2)) + ' $m.s^{-1}$ | $h_{0}$/c = ' + str(np.round(h0_c, 2)) + ' | $x_{f}$/c = ' + str(np.round(xp, 2)), fontsize=10)
    plt.legend(loc='lower right')

    plt.savefig('./files/aero/reduced_frequency_theta_0_graph/reduced_frequency_theta_0_graph_u' + str(np.round(u_inf, 2)) + '_hc' + str(np.round(h0_c, 2)) + '_xf' + str(np.round(xp, 2)) + '.png', bbox_inches='tight')

plt.show()
