import os
import numpy as np
import torch
from mpi4py import MPI
from utils import aeroelasticity, kinematics, mlp, controller


comm = MPI.COMM_WORLD
rank = comm.Get_rank()
mpi_size = comm.Get_size()

if rank == 0:
    print(os.getcwd())

# Kinematics
frequency_arr = np.arange(1.0, 8.5, 0.5)

pitching_amplitude = 79.5
rotation_axis_in_percentage_chord = 0.32

# Dynamical System
m = 1.9
K_theta = 0.0
C_theta = 0.0

params_path = 'aeroelasticity/files/params/'

mode = 'semi_active'

model_name = 'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_3'
inputs_variables = ["x_f", "f", "theta_0"]
outputs_variable = ["lift_coefficient", "moment_coefficient"]

n_input, n_output, layers_configuration = mlp.load_parameters(neural_networks_path='mlp/files/neural_networks/',
                                                              model_name=model_name)

model = mlp.MLP(n_input=n_input,
                n_output=n_output,
                layers_configuration=layers_configuration)

model.load_state_dict(torch.load('mlp/files/neural_networks/' + model_name + '/model_' + model_name + '.pt',
                                 weights_only=True))
model.eval()

if rank == 0:
    if not os.path.exists(params_path + model_name + '/'):
        os.mkdir(params_path + model_name + '/')

for i in range(np.size(frequency_arr, 0)):
    frequency = frequency_arr[i]

    save_path = params_path + model_name + '/' + 'f_' + str(np.round(frequency, 2)) + '/'

    kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                           nb_time_step_per_period=1000,
                                           input_variable_names=inputs_variables,
                                           output_variable_names=outputs_variable,
                                           variable_values=[rotation_axis_in_percentage_chord, frequency, pitching_amplitude],
                                           m=3,
                                           tau=9,
                                           nb_period=6)
    kinematics_obj.generate_kinematics()

    controller_obj = controller.PIDController(K_p=np.array([0.0, 500.0]),
                                              K_i=np.array([0.0, 75.0]),
                                              K_d=np.array([0.0, 10.0]),
                                              mode=mode)

    cases = np.array([[0.5, 40.0, 25], [5.0, 5000.0, 50]])

    aeroelasticity.parametric_study(cases=cases,
                                    kinematics_obj=kinematics_obj,
                                    model=model,
                                    controller_obj=controller_obj,
                                    m=m,
                                    K_theta=K_theta,
                                    C_theta=C_theta,
                                    comm=comm,
                                    mpi_size=mpi_size,
                                    rank=rank,
                                    save_path=save_path)

