import matplotlib.pyplot as plt
import torch
import tqdm

from utils import kinematics, ldvm, results, mlp, controller, aeroelasticity
import numpy as np

frequency = 5.812756
K_h = 2415.15685
C_h = 18.712705

# Kinematics
pitching_amplitude = 79.5
rotation_axis_in_percentage_chord = 0.32
chord = 0.25
u_inf = 10.0

# Dynamical System
m = 1.9
K_theta = 0.0
C_theta = 0.0
theta_0 = 79.5
xf_c = 0.32

reduced_frequency = ((2 * np.pi * frequency) * (chord / 2)) / u_inf

model_name = 'outputs_cl_cm_inputs_xf_f_theta_ref_1_refined_5'
inputs_variables = ["x_f", "f", "theta_0"]
outputs_variable = ["lift_coefficient", "moment_coefficient"]

n_input, n_output, layers_configuration = mlp.load_parameters(neural_networks_path='mlp/files/neural_networks/',
                                                              model_name=model_name)

model = mlp.MLP(n_input=n_input,
                n_output=n_output,
                layers_configuration=layers_configuration)

model.load_state_dict(torch.load('mlp/files/neural_networks/' + model_name + '/model_' + model_name + '.pt',
                                 weights_only=True))
model.eval()

kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                       nb_time_step_per_period=1000,
                                       input_variable_names=inputs_variables,
                                       output_variable_names=outputs_variable,
                                       variable_values=[rotation_axis_in_percentage_chord, frequency, pitching_amplitude],
                                       m=3,
                                       tau=9,
                                       nb_period=6)
kinematics_obj.generate_kinematics()

controller_obj = controller.PIDController(K_p=np.array([0.0, 500.0]),
                                          K_i=np.array([0.0, 75.0]),
                                          K_d=np.array([0.0, 10.0]),
                                          mode='semi_active')

dynamical_system_obj = aeroelasticity.DynamicalSystem(kinematics_obj=kinematics_obj,
                                                      model=model,
                                                      controller=controller_obj,
                                                      m=m,
                                                      K_h=K_h,
                                                      K_theta=K_theta,
                                                      C_h=C_h,
                                                      C_theta=C_theta)
try:
    dynamical_system_obj.motion_solver()
except:
    pass

kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                       nb_time_step_per_period=1000,
                                       input_variable_names=["x_f", "f", "theta_0"],
                                       output_variable_names=["lift_coefficient", "moment_coefficient"],
                                       variable_values=[rotation_axis_in_percentage_chord, frequency, pitching_amplitude],
                                       heaving_amplitude_chord=0.5,
                                       m=3,
                                       tau=9,
                                       input_time_arr=dynamical_system_obj.time_period,
                                       input_h_arr=dynamical_system_obj.h_period,
                                       input_theta_arr=dynamical_system_obj.theta_period)

solver_input_data_dict = {
    'model': 'LDVM',
    'movement': {"frequency": kinematics_obj.frequency},
    'name': 'ldvm',
    'solver': {'duration': 2.0 / kinematics_obj.frequency,
               'nb_time_steps': 500},
    'airfoil': {'aileron_dimension_in_percentage_chord': 0.0,
                'chord': kinematics_obj.chord,
                'critical_lesp': 0.14,
                'rotation_axis_in_percentage_chord': kinematics_obj.rotation_axis_in_percentage_chord},
    'experimental_data': {'drag_coefficient': None,
                          'lift_coefficient': None,
                          'moment_coefficient': None},
    'flow': {'freestream_velocity': kinematics_obj.freestream_velocity,
             'surge_amplitude': 0.0,
             'surge_frequency': 0.0}
}

results_obj = ldvm.run_case(solver_input_data_dict, kinematics_obj, save_visualisation=True)
