import os
import signal
import matplotlib.pyplot as plt
import numpy as np
import scipy
from utils import results, state_space


class DynamicalSystem:
    def __init__(self, kinematics_obj, model, controller, m, K_h, K_theta, C_h, C_theta):
        self.mode = controller.mode
        self.success = False

        self.m = m
        self.K_h = K_h
        self.K_theta = K_theta
        self.rho = 1.225
        self.b = 1.0
        self.C_h = C_h
        self.C_theta = C_theta

        self.kinematics_obj = kinematics_obj
        self.model = model
        self.controller = controller

        self.lift = None
        self.moment = None

        self.control_lift = None
        self.control_moment = None

        self.time = None
        self.time_solver = None
        self.x = None

        self.h = None
        self.h_dot = None
        self.theta = None
        self.theta_dot = None
        self.h_period = None
        self.h_dot_period = None
        self.theta_period = None
        self.theta_dot_period = None

        self.time_period = None
        self.h_dot_period = None
        self.theta_dot_period = None

        self.lift_period = None
        self.moment_period = None

        self.control_lift_period = None
        self.control_moment_period = None

        self.stiffness_h_force = None
        self.stiffness_theta_force = None
        self.damping_h_force = None
        self.damping_theta_force = None
        self.stiffness_h_force_period = None
        self.stiffness_theta_force_period = None
        self.damping_h_force_period = None
        self.damping_theta_force_period = None

        self.heaving_power = None
        self.pitching_power = None

        self.heaving_energy = None
        self.pitching_energy = None

        self.heaving_power_recoverable = None
        self.pitching_power_recoverable = None
        self.heaving_energy_recoverable = None
        self.pitching_energy_recoverable = None

        self.heaving_power_consumed = None
        self.pitching_power_consumed = None
        self.heaving_energy_consumed = None
        self.pitching_energy_consumed = None

        self.mean_heaving_power = None
        self.mean_pitching_power = None
        self.power_recoverable = None
        self.power_consumed = None

        self.energy_recoverable = None
        self.energy_consumed = None

        self.vertical_extent_airfoil_motion = None
        self.flow_power = None

        self.energy_extracted = None
        self.mean_power_extracted = None

        self.efficiency = 0.0

    def motion_solver(self, debug=False):
        state_space_func = state_space.state_space_func

        x_0 = [self.kinematics_obj.h[0], 0.0, self.kinematics_obj.theta[0], 0.0]

        time = self.kinematics_obj.time[((self.kinematics_obj.m - 1) * self.kinematics_obj.tau):]
        dt = self.kinematics_obj.dt

        # class TimeoutException(Exception):
        #     pass
        #
        # def timeout_handler(signum, frame):
        #     raise TimeoutException()
        #
        # signal.signal(signal.SIGALRM, timeout_handler)
        #
        # try:
        #     signal.alarm(90)

        result = scipy.integrate.solve_ivp(state_space_func,
                                           t_span=(time.min(), time.max()),
                                           y0=x_0,
                                           t_eval=time,
                                           args=(
                                           self.kinematics_obj, self.model, self.controller, time, self.m, self.b,
                                           self.C_h, self.C_theta, self.K_h, self.K_theta, self.rho),
                                           max_step=dt,
                                           atol=10000000.0,
                                           rtol=10000000.0,
                                           first_step=dt)

            # signal.alarm(0)

        # except TimeoutException:
        #     print('TIMED-OUT: f = ' + str(self.kinematics_obj.frequency) + ' | C_h = ' + str(self.C_h) + ' | K_h = ' + str(self.K_h))
        #     result = None

        # if result is not None:

        if debug:
            print(result)

        self.success = result.success

        if self.success:
            self.time = state_space_func.time
            self.time_solver = result.t
            self.x = result.y

            self.lift = np.interp(self.kinematics_obj.time, self.time, state_space_func.lift)
            self.moment = np.interp(self.kinematics_obj.time, self.time, state_space_func.moment)

            self.control_lift = np.interp(self.kinematics_obj.time, self.time, state_space_func.control_lift)
            self.control_moment = np.interp(self.kinematics_obj.time, self.time, state_space_func.control_moment)

            self.stiffness_h_force = np.interp(self.kinematics_obj.time, self.time, state_space_func.stiffness_h_force)
            self.stiffness_theta_force = np.interp(self.kinematics_obj.time, self.time, state_space_func.stiffness_theta_force)
            self.damping_h_force = np.interp(self.kinematics_obj.time, self.time, state_space_func.damping_h_force)
            self.damping_theta_force = np.interp(self.kinematics_obj.time, self.time, state_space_func.damping_theta_force)

            self.h = np.interp(self.kinematics_obj.time, self.time_solver, self.x[0])
            self.h_dot = np.interp(self.kinematics_obj.time, self.time_solver, self.x[1])
            self.theta = np.interp(self.kinematics_obj.time, self.time_solver, self.x[2])
            self.theta_dot = np.interp(self.kinematics_obj.time, self.time_solver, self.x[3])

            self.time = self.kinematics_obj.time

        self.compute_results()

    def compute_results(self):
        if self.success:
            self.extract_period()

            self.compute_pitching_power()
            self.compute_heaving_power()

            self.compute_flow_power()

            if self.mode == 'fully_active':
                self.energy_recoverable = self.pitching_energy_recoverable + self.heaving_energy_recoverable
                self.energy_consumed = self.pitching_energy_consumed + self.heaving_energy_consumed

            elif self.mode == 'semi_active':
                self.energy_recoverable = self.heaving_energy
                self.energy_consumed = self.pitching_energy

                self.power_recoverable = self.mean_heaving_power
                self.power_consumed = self.mean_pitching_power

            self.energy_extracted = (self.pitching_energy + self.heaving_energy)
            self.mean_power_extracted = (1.0 / self.kinematics_obj.period) * results.compute_integral(x=self.time_period,
                                                                                                      y=(self.pitching_power + self.heaving_power))

            self.efficiency = self.mean_power_extracted / self.flow_power
        else:
            self.efficiency = 0.0
            self.energy_recoverable = 0.0
            self.energy_consumed = 0.0
            self.power_recoverable = 0.0
            self.power_consumed = 0.0

    def compute_flow_power(self):
        self.vertical_extent_airfoil_motion = 2.0 * np.abs(self.h_period).max()
        self.flow_power = 0.5 * self.rho * (self.kinematics_obj.freestream_velocity ** 3) * self.vertical_extent_airfoil_motion * self.b

    def compute_heaving_power(self):
        if self.mode == 'fully_active':
            self.heaving_power = - self.control_lift_period * self.h_dot_period

        elif self.mode == 'semi_active':
            self.heaving_power = - self.damping_h_force_period * self.h_dot_period

        self.heaving_energy = results.compute_integral(x=self.time_period,
                                                       y=self.heaving_power)

        self.mean_heaving_power = self.kinematics_obj.frequency * self.heaving_energy

        self.heaving_power_recoverable = np.maximum(self.heaving_power, np.zeros(np.size(self.heaving_power, 0)))

        self.heaving_energy_recoverable = results.compute_integral(x=self.time_period,
                                                                   y=self.heaving_power_recoverable)

        self.heaving_power_consumed = np.minimum(self.heaving_power, np.zeros(np.size(self.heaving_power, 0)))

        self.heaving_energy_consumed = results.compute_integral(x=self.time_period,
                                                                y=self.heaving_power_consumed)

    def extract_period(self):
        self.time_period = self.time[-(self.kinematics_obj.nb_time_step_per_period + 1):]
        self.time_period = np.abs(self.time_period - self.time_period[0])
        self.h_period = self.h[-(self.kinematics_obj.nb_time_step_per_period + 1):]
        self.h_dot_period = self.h_dot[-(self.kinematics_obj.nb_time_step_per_period + 1):]
        self.theta_period = self.theta[-(self.kinematics_obj.nb_time_step_per_period + 1):]
        self.theta_dot_period = self.theta_dot[-(self.kinematics_obj.nb_time_step_per_period + 1):]

        self.lift_period = self.lift[-(self.kinematics_obj.nb_time_step_per_period + 1):]
        self.moment_period = self.moment[-(self.kinematics_obj.nb_time_step_per_period + 1):]

        self.control_lift_period = self.control_lift[-(self.kinematics_obj.nb_time_step_per_period + 1):]
        self.control_moment_period = self.control_moment[-(self.kinematics_obj.nb_time_step_per_period + 1):]

        self.stiffness_h_force_period = self.stiffness_h_force[-(self.kinematics_obj.nb_time_step_per_period + 1):]
        self.stiffness_theta_force_period = self.stiffness_theta_force[-(self.kinematics_obj.nb_time_step_per_period + 1):]

        self.damping_h_force_period = self.damping_h_force[-(self.kinematics_obj.nb_time_step_per_period + 1):]
        self.damping_theta_force_period = self.damping_theta_force[-(self.kinematics_obj.nb_time_step_per_period + 1):]

    def compute_pitching_power(self):
        if self.mode == 'fully_active':
            self.pitching_power = - self.control_moment_period * self.theta_dot_period
        elif self.mode == 'semi_active':
            self.pitching_power = - np.abs(self.control_moment_period * self.theta_dot_period)

        self.pitching_energy = results.compute_integral(x=self.time_period,
                                                        y=self.pitching_power)

        self.mean_pitching_power = self.kinematics_obj.frequency * self.pitching_energy

        self.pitching_power_recoverable = np.maximum(self.pitching_power, np.zeros(np.size(self.pitching_power, 0)))
        self.pitching_energy_recoverable = results.compute_integral(x=self.time_period,
                                                                    y=self.pitching_power_recoverable)
        self.pitching_power_consumed = np.minimum(self.pitching_power, np.zeros(np.size(self.pitching_power, 0)))
        self.pitching_energy_consumed = results.compute_integral(x=self.time_period,
                                                                 y=self.pitching_power_consumed)

    def show(self):
        if self.success:
            fig, axs = plt.subplots(1, 3, figsize=(16, 5))
            plt.subplots_adjust(left=0.1, bottom=0.1, right=0.9, top=0.9, wspace=0.5, hspace=0.2)

            time_ref = self.kinematics_obj.time[-self.kinematics_obj.nb_time_step_per_period:] - self.kinematics_obj.time[-self.kinematics_obj.nb_time_step_per_period]
            theta_ref = self.kinematics_obj.theta[-self.kinematics_obj.nb_time_step_per_period:]

            axs[0].plot(self.time_period * self.kinematics_obj.frequency, self.h_period / self.kinematics_obj.chord, label='$\\frac{h}{c}$', color='green')
            # axs[0].plot(self.kinematics_obj.time[-self.kinematics_obj.nb_time_step_per_period:], self.kinematics_obj.h[-self.kinematics_obj.nb_time_step_per_period:], linestyle='dashed', label='$h_{ref}$', color='green')

            axs_y2 = axs[0].twinx()
            axs_y2.set_ylabel('$\\theta$ ($^{\\circ}$)')
            axs_y2.plot(self.time_period * self.kinematics_obj.frequency, np.degrees(self.theta_period), label='$\\theta$', color='red')
            axs_y2.plot(time_ref * self.kinematics_obj.frequency, np.degrees(theta_ref), linestyle='dashed', label='$\\theta_{ref}$', color='red')
            axs_y2.legend(fontsize=9)

            axs[0].grid()
            axs[0].set_xlabel('$\\frac{t}{T}$')
            axs[0].set_ylabel('$\\frac{h}{c}$')
            axs[0].set_xlim([0.0, 1.0])
            axs[0].legend(fontsize=9, loc='upper center')

            axs[1].plot(self.time_period * self.kinematics_obj.frequency, self.lift_period, label='$L$', color='green')
            # axs[1].plot(self.time_period * self.kinematics_obj.frequency, self.control_lift_period, linestyle='dashed',
            #             label='$L_{control}$', color='green')

            axs_y3 = axs[1].twinx()
            axs_y3.set_ylabel('$M$ ($Nm.m^{-1}$)')
            axs_y3.plot(self.time_period * self.kinematics_obj.frequency, self.moment_period, label='$M$', color='red')
            axs_y3.plot(self.time_period * self.kinematics_obj.frequency, self.control_moment_period, linestyle='dashed',
                        label='$M_{control}$', color='red')
            axs_y3.legend(fontsize=9, loc='lower left')

            axs[1].grid()
            axs[1].set_xlabel('$\\frac{t}{T}$')
            axs[1].set_xlim([0.0, 1.0])
            axs[1].set_ylabel('$L$ ($N.m^{-1}$)')
            axs[1].set_ylim([-70.0, 70.0])
            axs_y3.set_ylim([-70.0, 70.0])
            axs[1].legend(fontsize=9, loc='upper right')

            axs[2].plot(self.time_period * self.kinematics_obj.frequency, self.heaving_power,
                        label='$P_{h}$ | $P_{h}$ = ' + str(np.round(self.mean_heaving_power, 2)) + ' $W.m^{-1}$', color='green')
            axs[2].plot(self.time_period * self.kinematics_obj.frequency, self.pitching_power,
                        label='$P_{\\theta}$ | $P_{\\theta}$ = ' + str(np.round(self.mean_pitching_power, 2)) + ' $W.m^{-1}$', color='red')
            axs[2].grid()
            axs[2].legend(fontsize=9)
            axs[2].set_xlabel('$\\frac{t}{T}$')
            axs[2].set_xlim([0.0, 1.0])
            axs[2].set_ylabel('Power ($W.m^{-1}$)')
            # axs[2].set_title('Energy Recoverable: ' + str(
            #     np.round(self.energy_recoverable, 3)) + ' J | Energy Consumed: ' + str(
            #     np.round(self.energy_consumed, 3)) + ' J', fontsize=9)

            reduced_frequency = ((2 * np.pi * self.kinematics_obj.frequency) * (self.kinematics_obj.chord / 2)) / self.kinematics_obj.freestream_velocity

            xf_c = self.kinematics_obj.rotation_axis_in_percentage_chord
            theta_0 = self.kinematics_obj.pitching_amplitude
            m = 1.9

            title = 'Step 5 | $U_{\\infty} = $' + str(self.kinematics_obj.freestream_velocity) + ' $m.s^{-1}$ | $m = ' + str(
                m) + '$ $kg.m^{-1}$' + " | $k = $" + str(np.round(reduced_frequency, 3)) + ' | $\\theta_0 = $' + str(
                np.round(np.degrees(theta_0), 2)) + '$^{\\circ}$ | $x_f/c = $' + str(xf_c) + ' | $C_h = $' + str(np.round(self.C_h, 3)) + ' $N.s.m^{-2}$ | $K_h = $' + str(np.round(self.K_h, 3)) + ' $N.m^{-2}$ | $\\eta = $' + str(
                np.round(self.efficiency, 3))

            plt.suptitle(title, fontsize=9.5)

            # plt.savefig('example_aeroelasticity.png', bbox_inches='tight')


def parametric_study(cases, kinematics_obj, model, controller_obj, m, K_theta, C_theta, comm, mpi_size, rank, save_path):
    if rank == 0:
        C_h_arr = np.linspace(cases[0][0], cases[0][1], int(cases[0][2]))
        K_h_arr = np.linspace(cases[1][0], cases[1][1], int(cases[1][2]))

        size = (np.size(C_h_arr, 0) * np.size(K_h_arr, 0))
        input_array = np.zeros((size, 2))
        efficiency = np.zeros(size)

        k = 0
        for i in range(np.size(C_h_arr, 0)):
            for j in range(np.size(K_h_arr, 0)):
                input_array[k, 0] = C_h_arr[i]
                input_array[k, 1] = K_h_arr[j]
                k = k + 1

        input_array = np.array_split(input_array, mpi_size)
        efficiency = np.array_split(efficiency, mpi_size)
    else:
        input_array = None
        efficiency = None

    input_array = comm.scatter(input_array, root=0)
    efficiency = comm.scatter(efficiency, root=0)

    for i in range(np.size(input_array, 0)):
        print('Process ' + str(rank + 1) + ': ' + str(i + 1) + '/' + str(np.size(input_array, 0)))

        inputs = input_array[i]
        C_h = inputs[0]
        K_h = inputs[1]

        dynamical_system_obj = DynamicalSystem(kinematics_obj=kinematics_obj,
                                               model=model,
                                               controller=controller_obj,
                                               m=m,
                                               K_h=K_h,
                                               K_theta=K_theta,
                                               C_h=C_h,
                                               C_theta=C_theta)

        try:
            dynamical_system_obj.motion_solver()
        except:
            pass

        if dynamical_system_obj.success:
            if np.abs(dynamical_system_obj.efficiency) < 2.0:
                efficiency[i] = dynamical_system_obj.efficiency
            else:
                efficiency[i] = 0.0
        else:
            efficiency[i] = 0.0

    input_array = comm.gather(input_array, root=0)
    efficiency = comm.gather(efficiency, root=0)

    if rank == 0:
        input_array = np.array(np.concatenate(input_array), dtype=float)
        efficiency = np.array(np.concatenate(efficiency), dtype=float)

        if not os.path.exists(save_path):
            os.mkdir(save_path)

        np.save(save_path + 'input_array.npy', input_array)
        np.save(save_path + 'efficiency.npy', efficiency)

        print("INFO: Parametric Study Saved!")
