import matplotlib.pyplot as plt
import numpy as np
from utils import solver, input_data, kinematics, results, controller, aeroelasticity, mlp
import torch
from mpi4py import MPI
import os


class Dataset(torch.utils.data.Dataset):
    # load the dataset
    def __init__(self, X, y):
        # store the inputs and outputs
        self.X = X
        self.y = y

    # number of rows in the dataset
    def __len__(self):
        return self.X.size()[0]

    # get a row at an index
    def __getitem__(self, idx):
        return [self.X[idx], self.y[idx]]


def generate_array_cases(range_variables):
    temp = []
    nb_variables = np.size(range_variables, 0)
    for i in range(nb_variables):
        temp.append(np.linspace(range_variables[i][0], range_variables[i][1], int(range_variables[i][2])))

    cases = []

    if nb_variables == 2:
        for k in range(np.size(temp[0])):
            var_1 = temp[0][k]
            for j in range(np.size(temp[1])):
                var_2 = temp[1][j]
                cases.append(np.array([var_1, var_2]))
    elif nb_variables == 3:
        for k in range(np.size(temp[0])):
            var_1 = temp[0][k]
            for j in range(np.size(temp[1])):
                var_2 = temp[1][j]
                for p in range(np.size(temp[2])):
                    var_3 = temp[2][p]
                    cases.append(np.array([var_1, var_2, var_3]))

    cases = np.array(cases)
    np.random.shuffle(cases)

    return cases


def initialisation(nb_time_steps_dataset, range_variables, rank, mpi_size, inputs_variables, outputs_variable):
    if rank == 0:
        cases_array_temp = generate_array_cases(range_variables)
        cases_array = np.array_split(cases_array_temp, mpi_size)
        cases_array = np.array(cases_array, dtype=object)

        dataset_inputs_temp = []
        dataset_outputs_temp = []

        for k in range(np.size(cases_array, 0)):
            dataset_inputs_temp.append(np.zeros((nb_time_steps_dataset * np.size(cases_array[k], 0), (len(inputs_variables) * 4 + 2))))
            dataset_outputs_temp.append(np.zeros((nb_time_steps_dataset * np.size(cases_array[k], 0),
                                                  len(outputs_variable))))

        dataset_inputs = np.array(dataset_inputs_temp, dtype=object)
        dataset_outputs = np.array(dataset_outputs_temp, dtype=object)
    else:
        cases_array = None
        dataset_inputs = None
        dataset_outputs = None

    return cases_array, dataset_inputs, dataset_outputs


def extract_period(results_obj, nb_time_steps):
    time_before = results_obj.time[results_obj.time >= results_obj.period] - results_obj.period
    theta = results_obj.theta[results_obj.time >= results_obj.period]
    theta_dot = results_obj.theta_dot[results_obj.time >= results_obj.period]
    theta_dot_2 = results_obj.theta_dot_2[results_obj.time >= results_obj.period]
    h = results_obj.h[results_obj.time >= results_obj.period]
    h_dot = results_obj.h_dot[results_obj.time >= results_obj.period]
    h_dot_2 = results_obj.h_dot_2[results_obj.time >= results_obj.period]
    lift_coefficient = results_obj.lift_coefficient[results_obj.time >= results_obj.period]
    moment_coefficient = results_obj.moment_airfoil_coefficient[results_obj.time >= results_obj.period]

    time, theta = results.interp_arr(time_before, theta, nb_time_steps)
    time, theta_dot = results.interp_arr(time_before, theta_dot, nb_time_steps)
    time, theta_dot_2 = results.interp_arr(time_before, theta_dot_2, nb_time_steps)
    time, h = results.interp_arr(time_before, h, nb_time_steps)
    time, h_dot = results.interp_arr(time_before, h_dot, nb_time_steps)
    time, h_dot_2 = results.interp_arr(time_before, h_dot_2, nb_time_steps)
    time, lift_coefficient = results.interp_arr(time_before, lift_coefficient, nb_time_steps)
    time, moment_coefficient = results.interp_arr(time_before, moment_coefficient, nb_time_steps)

    return time, theta, theta_dot, theta_dot_2, h, h_dot, h_dot_2, lift_coefficient, moment_coefficient


def get_dataset_case(results_obj, kinematics_obj, nb_time_steps, debug):
    (time, theta, theta_dot, theta_dot_2, h, h_dot, h_dot_2, lift_coefficient, moment_coefficient)\
        = extract_period(results_obj, nb_time_steps)

    efficiency = results_obj.efficiency

    kinematics_obj.set_results(time, theta, theta_dot, theta_dot_2, h, h_dot, h_dot_2, lift_coefficient,
                               moment_coefficient)

    if debug:
        plot_results(kinematics_obj, lift_coefficient, moment_coefficient)

    inputs, outputs = kinematics_obj.get_lines()

    return inputs, outputs


def run_case(inputs_obj):
    solver_input_data_dict = {
        'model': 'LDVM',
        'movement': {"frequency": inputs_obj.frequency},
        'name': 'ldvm',
        'solver': {'duration': 2.0 / inputs_obj.frequency,
                   'nb_time_steps': 500},
        'airfoil': {'aileron_dimension_in_percentage_chord': 0.0,
                                             'chord': inputs_obj.chord,
                                             'critical_lesp': 0.14,
                                             'rotation_axis_in_percentage_chord': inputs_obj.rotation_axis_in_percentage_chord},
                                 'experimental_data': {'drag_coefficient': None,
                                                       'lift_coefficient': None,
                                                       'moment_coefficient': None},
                                 'flow': {'freestream_velocity': inputs_obj.freestream_velocity,
                                          'surge_amplitude': 0.0,
                                          'surge_frequency': 0.0}
        }

    solver_input_data_obj = input_data.InputData(solver_input_data_dict)

    solver_obj = solver.Solver(initial_data=[solver_input_data_obj])
    solver_obj.solve(func=inputs_obj.kinematics_function)
    results_obj = solver_obj.output()

    return results_obj


def run_case_with_dynamical_system_predicting_motion(model, inputs_variables, outputs_variable, frequency, C_h, K_h):
    # Kinematics
    pitching_amplitude = 79.5
    rotation_axis_in_percentage_chord = 0.32

    # Dynamical System
    m = 1.9
    K_theta = 0.0
    C_theta = 0.0

    kinematics_obj_for_dyn = kinematics.Kinematics(kinematics_reference=1,
                                                   nb_time_step_per_period=1000,
                                                   input_variable_names=inputs_variables,
                                                   output_variable_names=outputs_variable,
                                                   variable_values=[rotation_axis_in_percentage_chord, frequency,
                                                                    pitching_amplitude],
                                                   m=3,
                                                   tau=9,
                                                   nb_period=6)

    kinematics_obj_for_dyn.generate_kinematics()

    controller_obj = controller.PIDController(K_p=np.array([0.0, 500.0]),
                                              K_i=np.array([0.0, 75.0]),
                                              K_d=np.array([0.0, 10.0]),
                                              mode='semi_active')

    dynamical_system_obj = aeroelasticity.DynamicalSystem(kinematics_obj=kinematics_obj_for_dyn,
                                                          model=model,
                                                          controller=controller_obj,
                                                          m=m,
                                                          K_h=K_h,
                                                          K_theta=K_theta,
                                                          C_h=C_h,
                                                          C_theta=C_theta)
    try:
        dynamical_system_obj.motion_solver()
    except:
        pass

    kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                           nb_time_step_per_period=1000,
                                           input_variable_names=inputs_variables,
                                           output_variable_names=outputs_variable,
                                           variable_values=[rotation_axis_in_percentage_chord, frequency,
                                                            pitching_amplitude],
                                           heaving_amplitude_chord=0.5,
                                           m=3,
                                           tau=9,
                                           input_time_arr=dynamical_system_obj.time_period,
                                           input_h_arr=dynamical_system_obj.h_period,
                                           input_theta_arr=dynamical_system_obj.theta_period)

    solver_input_data_dict = {
        'model': 'LDVM',
        'movement': {"frequency": kinematics_obj.frequency},
        'name': 'ldvm',
        'solver': {'duration': 2.0 / kinematics_obj.frequency,
                   'nb_time_steps': 500},
        'airfoil': {'aileron_dimension_in_percentage_chord': 0.0,
                    'chord': kinematics_obj.chord,
                    'critical_lesp': 0.14,
                    'rotation_axis_in_percentage_chord': kinematics_obj.rotation_axis_in_percentage_chord},
        'experimental_data': {'drag_coefficient': None,
                              'lift_coefficient': None,
                              'moment_coefficient': None},
        'flow': {'freestream_velocity': kinematics_obj.freestream_velocity,
                 'surge_amplitude': 0.0,
                 'surge_frequency': 0.0}
    }

    solver_input_data_obj = input_data.InputData(solver_input_data_dict)

    solver_obj = solver.Solver(initial_data=[solver_input_data_obj])
    solver_obj.solve(func=kinematics_obj.kinematics_function)
    results_obj = solver_obj.output()

    return kinematics_obj, results_obj

def plot_results(kinematics_obj, lift_coefficient, moment_coefficient):
    fig, axs = plt.subplots(1, 3, figsize=(12, 5))

    if kinematics_obj.input_time_arr is not None and kinematics_obj.input_h_arr is not None and kinematics_obj.input_theta_arr is not None:
        axs[0].plot(kinematics_obj.time, kinematics_obj.theta, label='$\\theta$')
        axs[0].plot(kinematics_obj.input_time_arr, kinematics_obj.input_theta_arr, label='$\\theta$', linestyle='dashed')
        axs[0].plot(kinematics_obj.time, kinematics_obj.h, label='h')
        axs[0].plot(kinematics_obj.input_time_arr, kinematics_obj.input_h_arr, label='h', linestyle='dashed')
        axs[0].set_xlabel('Time (s)')
        axs[0].grid()
        axs[0].legend()

        axs[1].plot(lift_coefficient)
        axs[1].plot(moment_coefficient)
        axs[1].grid()
    else:
        axs[0].plot(kinematics_obj.time, kinematics_obj.theta, label='$\\theta$')
        axs[0].plot(kinematics_obj.time, kinematics_obj.h, label='h')
        axs[0].set_xlabel('Time (s)')
        axs[0].grid()
        axs[0].legend()

        axs[1].plot(lift_coefficient)
        axs[1].plot(moment_coefficient)
        axs[1].grid()

    rho = 1.225

    dynamic_pressure = (1 / 2) * rho * (kinematics_obj.freestream_velocity ** 2)
    lift = lift_coefficient * dynamic_pressure * kinematics_obj.chord
    moment = moment_coefficient * dynamic_pressure * (kinematics_obj.chord ** 2)

    heaving_power = lift * kinematics_obj.h_dot
    pitching_power = moment * kinematics_obj.theta_dot
    pitching_energy = (kinematics_obj.time[-1] - kinematics_obj.time[0]) * np.mean(pitching_power[~np.isnan(pitching_power)])
    heaving_energy = (kinematics_obj.time[-1] - kinematics_obj.time[0]) * np.mean(heaving_power[~np.isnan(heaving_power)])

    axs[2].plot(lift_coefficient * kinematics_obj.h_dot, label='$P_{h}$ | $E_{h}$ = ' + str(np.round(heaving_energy, 4)) + ' J.')
    axs[2].plot(moment_coefficient * kinematics_obj.theta_dot, label='$P_{\\theta}$ | $E_{\\theta}$ = ' + str(np.round(pitching_energy, 4)) + ' J.')
    axs[2].grid()
    axs[2].legend()

    plt.suptitle("$\\theta_{0} = $" + str(kinematics_obj.pitching_amplitude) + " $^{\\circ}$ | $U_{inf} = $" + str(kinematics_obj.freestream_velocity) + " | $f = $" + str(kinematics_obj.frequency), fontsize=10.5)

    plt.show()


def iterations(nb_time_steps, nb_time_steps_dataset, inputs_variables, outputs_variable, cases, dataset_inputs, dataset_outputs, rank,
               m, tau, heaving_amplitude_chord, kinematics_reference, debug, model):
    k = np.arange(0, np.size(cases, 0))
    np.random.shuffle(k)
    k = k * nb_time_steps_dataset

    for i in range(np.size(cases, 0)):
        print('Process ' + str(rank + 1) + ': ' + str(i + 1) + '/' + str(np.size(cases, 0)))

        if model is None:
            kinematics_obj = kinematics.Kinematics(kinematics_reference=kinematics_reference,
                                                   nb_time_step_per_period=nb_time_steps,
                                                   input_variable_names=inputs_variables,
                                                   output_variable_names=outputs_variable,
                                                   variable_values=cases[i],
                                                   heaving_amplitude_chord=heaving_amplitude_chord,
                                                   m=m,
                                                   tau=tau)
            results_obj = run_case(kinematics_obj)
        else:
            case = cases[i]

            frequency = case[0]
            C_h = case[1]
            K_h = case[2]

            kinematics_obj, results_obj = run_case_with_dynamical_system_predicting_motion(model, inputs_variables, outputs_variable, frequency, C_h, K_h)

        inputs, outputs = get_dataset_case(results_obj, kinematics_obj, nb_time_steps, debug)

        for nb in range(len(inputs)):
            dataset_inputs[k[i] + nb] = inputs[nb]
            dataset_outputs[k[i] + nb] = outputs[nb]

    return dataset_inputs, dataset_outputs


def iterations_refined(nb_time_steps, nb_time_steps_dataset, inputs_variables, outputs_variable, cases, dataset_inputs, dataset_outputs, rank,
                       m, tau, heaving_amplitude_chord, kinematics_reference, debug):
    k = np.arange(0, np.size(cases, 0))
    np.random.shuffle(k)
    k = k * nb_time_steps_dataset

    for i in range(np.size(cases, 0)):
        print('Process ' + str(rank + 1) + ': ' + str(i + 1) + '/' + str(np.size(cases, 0)))

        kinematics_obj = kinematics.Kinematics(kinematics_reference=kinematics_reference,
                                               nb_time_step_per_period=nb_time_steps,
                                               input_variable_names=inputs_variables,
                                               output_variable_names=outputs_variable,
                                               variable_values=cases[i],
                                               heaving_amplitude_chord=heaving_amplitude_chord,
                                               m=m,
                                               tau=tau)
        results_obj = run_case(kinematics_obj)

        inputs, outputs = get_dataset_case(results_obj, kinematics_obj, nb_time_steps, debug)

        for nb in range(len(inputs)):
            dataset_inputs[k[i] + nb] = inputs[nb]
            dataset_outputs[k[i] + nb] = outputs[nb]

    return dataset_inputs, dataset_outputs


def generate(nb_time_steps, nb_time_steps_dataset, inputs_variables, outputs_variable, range_variables, nb_cases_train_dataset,
             nb_cases_test_dataset, nb_cases_validation_dataset, dataset_name, m, tau, heaving_amplitude_chord=0.5, kinematics_reference=1,
             refined=False,save=True, debug=False, model=None):
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    mpi_size = comm.Get_size()

    if rank == 0:
        print(os.getcwd())

    cases_array, dataset_inputs, dataset_outputs = initialisation(nb_time_steps_dataset, range_variables, rank,
                                                                  mpi_size, inputs_variables, outputs_variable)

    cases_array = comm.scatter(cases_array, root=0)
    dataset_inputs = comm.scatter(dataset_inputs, root=0)
    dataset_outputs = comm.scatter(dataset_outputs, root=0)

    dataset_inputs, dataset_outputs = iterations(nb_time_steps,
                                                 nb_time_steps_dataset, inputs_variables, outputs_variable, cases_array,
                                                 dataset_inputs, dataset_outputs, rank, m, tau, heaving_amplitude_chord,
                                                 kinematics_reference, debug, model)

    cases_array = comm.gather(cases_array, root=0)
    dataset_inputs = comm.gather(dataset_inputs, root=0)
    dataset_outputs = comm.gather(dataset_outputs, root=0)

    if rank == 0:
        dataset_inputs = np.array(np.concatenate(dataset_inputs), dtype=float)
        dataset_outputs = np.array(np.concatenate(dataset_outputs), dtype=float)

        train_dataset, test_dataset, validation_dataset = split_dataset(dataset_inputs, dataset_outputs,
                                                                        nb_time_steps_dataset,
                                                                        nb_cases_train_dataset,
                                                                        nb_cases_test_dataset,
                                                                        nb_cases_validation_dataset)

        if save is True:
            save_dataset(dataset_name, train_dataset, test_dataset, validation_dataset)
        else:
            pass


def split_dataset(dataset_inputs, dataset_outputs, nb_time_steps_dataset, nb_cases_train_dataset,
                  nb_cases_test_dataset, nb_cases_validation_dataset):
    index_train = nb_time_steps_dataset * nb_cases_train_dataset
    index_test = nb_time_steps_dataset * (nb_cases_train_dataset + nb_cases_test_dataset)
    index_validation = nb_time_steps_dataset * (nb_cases_train_dataset + nb_cases_test_dataset + nb_cases_validation_dataset)

    train_dataset = Dataset(torch.tensor(dataset_inputs[:index_train]),
                            torch.tensor(dataset_outputs[:index_train]))

    test_dataset = Dataset(torch.tensor(dataset_inputs[index_train:index_test]),
                           torch.tensor(dataset_outputs[index_train:index_test]))

    validation_dataset = Dataset(torch.tensor(dataset_inputs[index_test:index_validation]),
                                 torch.tensor(dataset_outputs[index_test:index_validation]))

    return train_dataset, test_dataset, validation_dataset

def save_dataset(dataset_name, train_dataset, test_dataset, validation_dataset):
    save_path = 'mlp/files/datasets/' + dataset_name + '/'
    if not os.path.exists(save_path):
        os.mkdir(save_path)
    torch.save(train_dataset, save_path + 'dataset_train.pt')
    torch.save(test_dataset, save_path + 'dataset_test.pt')
    torch.save(validation_dataset, save_path + 'dataset_validation.pt')
    print("INFO: Dataset Saved!")