import numpy as np
import torch
from utils import results


class KinematicsFunctionBuilder:
    def __init__(self, reference, frequency, pitching_amplitude, heaving_amplitude, input_time_arr=None,
                 input_h_arr=None, input_theta_arr=None, beta=None):
        self.reference = int(reference)

        self.frequency = frequency
        self.pitching_amplitude = pitching_amplitude
        self.heaving_amplitude = heaving_amplitude
        self.beta = beta

        if input_time_arr is not None and input_h_arr is not None and input_theta_arr is not None:
            self.input_time_arr = input_time_arr
            self.input_h_arr = input_h_arr
            self.input_theta_arr = input_theta_arr
        else:
            self.input_time_arr = None
            self.input_h_arr = None
            self.input_theta_arr = None

    def get_kinematics_function(self):

        if self.input_time_arr is None and self.input_h_arr is None and self.input_theta_arr is None:

            if self.reference == 1:

                def kinematics_function(time):
                    self.angular_pulsation = 2.0 * np.pi * self.frequency

                    theta = self.pitching_amplitude * np.cos(self.angular_pulsation * time)
                    h = self.heaving_amplitude * np.sin(self.angular_pulsation * time)

                    return theta, h

                return kinematics_function

            elif self.reference == 2:

                def kinematics_function(time):
                    self.angular_pulsation = 2.0 * np.pi * self.frequency

                    h = self.heaving_amplitude * np.sin(self.angular_pulsation * time)

                    if not isinstance(time, np.ndarray):
                        time_mod = np.mod(time, 1.0 / self.frequency)

                        if 0 <= time_mod <= ((np.pi / (2 * self.angular_pulsation)) * (1 - 1 / self.beta)):
                            theta = self.pitching_amplitude

                        elif (np.pi / (2 * self.angular_pulsation)) * (1 - 1 / self.beta) < time_mod <= (np.pi / (2 * self.angular_pulsation)) * (1 + 1 / self.beta):
                            theta = self.pitching_amplitude * np.sin(self.beta * self.angular_pulsation * time_mod + np.pi * (1 - self.beta / 2))

                        elif (np.pi / (2 * self.angular_pulsation)) * (1 + 1 / self.beta) < time_mod <= (np.pi / (2 * self.angular_pulsation)) * (3 - 1 / self.beta):
                            theta =  - self.pitching_amplitude

                        elif (np.pi / (2 * self.angular_pulsation)) * (3 - 1 / self.beta) < time_mod <= (np.pi / (2 * self.angular_pulsation)) * (3 + 1 / self.beta):
                            theta = self.pitching_amplitude * np.sin(self.beta * self.angular_pulsation * time_mod + np.pi * (2 - 3 * self.beta / 2))

                        elif (np.pi / (2 * self.angular_pulsation)) * (3 + 1 / self.beta) < time_mod <= (2 * np.pi / self.angular_pulsation):
                            theta = self.pitching_amplitude
                        else:
                            raise Exception
                    else:
                        theta = np.zeros(np.size(time, 0))
                        for i in range(np.size(time, 0)):
                            time_mod = np.mod(time[i], 1.0 / self.frequency)

                            if 0 <= time_mod <= ((np.pi / (2 * self.angular_pulsation)) * (1 - 1 / self.beta)):
                                theta[i] = self.pitching_amplitude

                            elif (np.pi / (2 * self.angular_pulsation)) * (1 - 1 / self.beta) < time_mod <= (
                                    np.pi / (2 * self.angular_pulsation)) * (1 + 1 / self.beta):
                                theta[i] = self.pitching_amplitude * np.sin(
                                    self.beta * self.angular_pulsation * time_mod + np.pi * (1 - self.beta / 2))

                            elif (np.pi / (2 * self.angular_pulsation)) * (1 + 1 / self.beta) < time_mod <= (
                                    np.pi / (2 * self.angular_pulsation)) * (3 - 1 / self.beta):
                                theta[i] = - self.pitching_amplitude

                            elif (np.pi / (2 * self.angular_pulsation)) * (3 - 1 / self.beta) < time_mod <= (
                                    np.pi / (2 * self.angular_pulsation)) * (3 + 1 / self.beta):
                                theta[i] = self.pitching_amplitude * np.sin(
                                    self.beta * self.angular_pulsation * time_mod + np.pi * (2 - 3 * self.beta / 2))

                            elif (np.pi / (2 * self.angular_pulsation)) * (3 + 1 / self.beta) < time_mod <= (
                                    2 * np.pi / self.angular_pulsation):
                                theta[i] = self.pitching_amplitude
                            else:
                                raise Exception

                    return theta, h

                return kinematics_function

            else:
                raise Exception('Wrong Kinematics Reference')

        elif self.input_time_arr is not None and self.input_h_arr is not None and self.input_theta_arr is not None:

            if self.reference == 1:

                def kinematics_function(time):
                    if np.round(time, 8) > 0.0 and np.mod(time, 1.0 / self.frequency) == 0.0:
                        time_mod = 1.0 / self.frequency
                    else:
                        time_mod = np.mod(time, 1.0 / self.frequency)

                    current_time_index = np.abs(self.input_time_arr - time_mod).argmin()

                    theta = self.input_theta_arr[current_time_index]
                    h = self.input_h_arr[current_time_index]

                    return theta, h

                return kinematics_function

            else:
                raise Exception('Wrong Kinematics Reference')


class Kinematics:
    def __init__(self, kinematics_reference, nb_time_step_per_period, input_variable_names, output_variable_names,
                 variable_values, heaving_amplitude_chord=0.5, m=3, tau=9, nb_period=1, input_time_arr=None,
                 input_h_arr=None, input_theta_arr=None):

        self.kinematics_reference = kinematics_reference

        # Default values
        self.chord = 0.25
        self.rotation_axis_in_percentage_chord = 0.3
        self.freestream_velocity = 10.0
        self.heaving_amplitude_chord = heaving_amplitude_chord

        self.nb_period = nb_period
        self.nb_time_step_per_period = nb_time_step_per_period

        self.input_variable_names = input_variable_names
        self.output_variable_names = output_variable_names
        self.kinematics_function = None
        self.frequency = None
        self.m = m
        self.tau = tau

        if input_h_arr is not None and input_time_arr is not None and input_theta_arr is not None:
            self.input_time_arr, self.input_h_arr = results.interp_arr(input_time_arr,
                                                                       input_h_arr,
                                                                       self.nb_time_step_per_period)
            self.input_time_arr, self.input_theta_arr = results.interp_arr(input_time_arr,
                                                                           input_theta_arr,
                                                                           self.nb_time_step_per_period)
        else:
            self.input_time_arr = None
            self.input_theta_arr = None
            self.input_h_arr = None

        self.time = None
        self.dt = None
        self.theta = None
        self.theta_dot = None
        self.theta_dot_2 = None
        self.h = None
        self.h_dot = None
        self.h_dot_2 = None
        self.lift_coefficient = None
        self.moment_coefficient = None

        if self.input_variable_names == ["theta_0", "h_0"]:
            self.frequency = 1.0
            self.pitching_amplitude = np.radians(variable_values[0])
            self.heaving_amplitude = variable_values[1]

            kinematics_function_builder = KinematicsFunctionBuilder(reference=self.kinematics_reference,
                                                                    frequency=self.frequency,
                                                                    pitching_amplitude=self.pitching_amplitude,
                                                                    heaving_amplitude=self.heaving_amplitude)

            self.kinematics_function = kinematics_function_builder.get_kinematics_function()

        elif self.input_variable_names == ["theta_0", "frequency"]:
            self.pitching_amplitude = np.radians(variable_values[0])
            self.frequency = variable_values[1]
            self.heaving_amplitude = self.heaving_amplitude_chord * self.chord

            kinematics_function_builder = KinematicsFunctionBuilder(reference=self.kinematics_reference,
                                                                    frequency=self.frequency,
                                                                    pitching_amplitude=self.pitching_amplitude,
                                                                    heaving_amplitude=self.heaving_amplitude)

            self.kinematics_function = kinematics_function_builder.get_kinematics_function()

        elif self.input_variable_names == ["u", "f", "theta_0"]:
            self.freestream_velocity = variable_values[0]
            self.pitching_amplitude = np.radians(variable_values[2])
            self.frequency = variable_values[1]
            self.heaving_amplitude = self.heaving_amplitude_chord * self.chord

            kinematics_function_builder = KinematicsFunctionBuilder(reference=self.kinematics_reference,
                                                                    frequency=self.frequency,
                                                                    pitching_amplitude=self.pitching_amplitude,
                                                                    heaving_amplitude=self.heaving_amplitude,
                                                                    beta=1.5)

            self.kinematics_function = kinematics_function_builder.get_kinematics_function()

        elif self.input_variable_names == ["x_f", "f", "theta_0"]:
            self.freestream_velocity = 10.0
            self.pitching_amplitude = np.radians(variable_values[2])
            self.frequency = variable_values[1]
            self.heaving_amplitude = self.heaving_amplitude_chord * self.chord
            self.rotation_axis_in_percentage_chord = variable_values[0]

            kinematics_function_builder = KinematicsFunctionBuilder(reference=self.kinematics_reference,
                                                                    frequency=self.frequency,
                                                                    pitching_amplitude=self.pitching_amplitude,
                                                                    heaving_amplitude=self.heaving_amplitude,
                                                                    beta=1.5,
                                                                    input_time_arr=self.input_time_arr,
                                                                    input_h_arr=self.input_h_arr,
                                                                    input_theta_arr=self.input_theta_arr)

            self.kinematics_function = kinematics_function_builder.get_kinematics_function()

        else:
            raise Exception

        self.period = 1.0 / self.frequency
        self.angular_frequency = self.frequency * 2.0 * np.pi
        self.feathering_parameter = self.pitching_amplitude / np.arctan((self.heaving_amplitude * self.angular_frequency)
                                                                        / self.freestream_velocity)
        self.reduced_frequency = (self.angular_frequency * (self.chord / 2.0)) / self.freestream_velocity

    def generate_kinematics(self):
        time = np.linspace(0.0, self.period * self.nb_period, self.nb_time_step_per_period * self.nb_period)
        self.dt = np.mean(np.diff(time))
        self.time = np.linspace(- self.dt * (self.m - 1) * self.tau, self.period * self.nb_period, int(self.nb_time_step_per_period * self.nb_period + (self.m - 1) * self.tau))
        self.theta, self.h = self.kinematics_function(self.time)
        self.theta_dot = np.insert(np.diff(self.theta) / self.dt, 0, np.nan)
        self.h_dot = np.insert(np.diff(self.h) / self.dt, 0, np.nan)

    def get_inputs_mlp(self):
        self.generate_kinematics()

        inputs_mlp = self.get_lines(without_outputs=True)

        self.time = self.time[18:]
        self.theta = self.theta[18:]
        self.h = self.h[18:]
        self.theta_dot = self.theta_dot[18:]
        self.h_dot = self.h_dot[18:]

        return torch.tensor(np.array(inputs_mlp, dtype=np.float32))

    def set_results(self, time, theta, theta_dot, theta_dot_2, h, h_dot, h_dot_2, lift_coefficient, moment_coefficient):
        self.time = time
        self.theta = theta
        self.theta_dot = theta_dot
        self.theta_dot_2 = theta_dot_2
        self.h = h
        self.h_dot = h_dot
        self.h_dot_2 = h_dot_2
        self.lift_coefficient = lift_coefficient
        self.moment_coefficient = moment_coefficient

    def get_line(self, i, without_outputs=False):
        if self.input_variable_names == ["theta_0", "h_0"] and self.output_variable_names == ["lift_coefficient"]:
            res_1 = np.array([self.theta[i], self.theta_dot[i], self.h[i], self.h_dot[i]])
            res_2 = np.array([self.theta[i - 9], self.theta_dot[i - 9], self.h[i - 9], self.h_dot[i - 9]])
            res_3 = np.array([self.theta[i - 18], self.theta_dot[i - 18], self.h[i - 18], self.h_dot[i - 18]])

            if not without_outputs:
                outputs_temp = self.lift_coefficient[i]

            inputs_temp = np.array([res_1, res_2, res_3]).flatten()

        elif self.input_variable_names == ["u", "f", "theta_0"] and self.output_variable_names == ["lift_coefficient",
                                                                                       "moment_coefficient"]:
            res_0 = np.array([self.freestream_velocity, self.frequency])
            res_1 = np.array([self.theta[i], self.theta_dot[i]])
            res_2 = np.array([self.h[i], self.h_dot[i]])
            res_3 = np.array([self.theta[i - int((self.m - 2.0) * self.tau)], self.theta_dot[i - int((self.m - 2.0) * self.tau)]])
            res_4 = np.array([self.h[i - int((self.m - 2.0) * self.tau)], self.h_dot[i - int((self.m - 2.0) * self.tau)]])
            res_5 = np.array([self.theta[i - int((self.m - 1.0) * self.tau)], self.theta_dot[i - int((self.m - 1.0) * self.tau)]])
            res_6 = np.array([self.h[i - int((self.m - 1.0) * self.tau)], self.h_dot[i - int((self.m - 1.0) * self.tau)]])

            if not without_outputs:
                outputs_temp = np.array([self.lift_coefficient[i], self.moment_coefficient[i]])

            inputs_temp = np.array([res_0, res_1, res_2, res_3, res_4, res_5, res_6]).flatten()

        elif self.input_variable_names == ["x_f", "f", "theta_0"] and self.output_variable_names == ["lift_coefficient",
                                                                                       "moment_coefficient"]:
            res_0 = np.array([self.rotation_axis_in_percentage_chord, self.frequency])
            res_1 = np.array([self.theta[i], self.theta_dot[i]])
            res_2 = np.array([self.h[i], self.h_dot[i]])
            res_3 = np.array([self.theta[i - int((self.m - 2.0) * self.tau)], self.theta_dot[i - int((self.m - 2.0) * self.tau)]])
            res_4 = np.array([self.h[i - int((self.m - 2.0) * self.tau)], self.h_dot[i - int((self.m - 2.0) * self.tau)]])
            res_5 = np.array([self.theta[i - int((self.m - 1.0) * self.tau)], self.theta_dot[i - int((self.m - 1.0) * self.tau)]])
            res_6 = np.array([self.h[i - int((self.m - 1.0) * self.tau)], self.h_dot[i - int((self.m - 1.0) * self.tau)]])

            if not without_outputs:
                outputs_temp = np.array([self.lift_coefficient[i], self.moment_coefficient[i]])

            inputs_temp = np.array([res_0, res_1, res_2, res_3, res_4, res_5, res_6]).flatten()

        else:
            raise Exception

        if not without_outputs:
            return inputs_temp, outputs_temp
        else:
            return inputs_temp

    def get_lines(self, without_outputs=False):
        inputs = []

        if not without_outputs:
            outputs = []

            for i in range(int((self.m - 1.0) * self.tau), np.size(self.time)):
                inputs_temp, outputs_temp = self.get_line(i)
                if np.all(np.isnan(outputs_temp) == False) and np.all(np.abs(outputs_temp) < 100.0):
                    inputs.append(inputs_temp)
                    outputs.append(outputs_temp)
                else:
                    inputs.append(np.zeros((1, (len(self.input_variable_names) * 4 + 2))))
                    outputs.append(np.zeros((1, len(self.output_variable_names))))

            inputs = np.array(inputs, dtype=object)
            outputs = np.array(outputs, dtype=object)

            if np.max(np.abs(self.lift_coefficient)) > 5.0 and np.max(np.abs(self.moment_coefficient)) > 5.0:
                inputs = np.array(np.zeros(np.shape(inputs)), dtype=object)
                outputs = np.array(np.zeros(np.shape(outputs)), dtype=object)

            return inputs, outputs
        else:
            for i in range(int((self.m - 1.0) * self.tau), np.size(self.time)):
                inputs_temp = self.get_line(i, without_outputs)
                inputs.append(inputs_temp)

            inputs = np.array(inputs, dtype=object)

            return inputs