import sys
from utils import results
import numpy as np
import sklearn
import torch
import matplotlib.pyplot as plt
import os
import tqdm


class MLP(torch.nn.Module):
    def __init__(self, n_input, n_output, layers_configuration):
        super().__init__()

        self.n_input = n_input
        self.n_output = n_output
        self.layers_configuration = layers_configuration

        self.layers = None
        self.build_layers()

    def build_layers(self):
        self.layers = torch.nn.Sequential()
        self.layers.append(torch.nn.Linear(self.n_input, self.layers_configuration[0]))
        self.layers.append(torch.nn.ReLU())
        for layer in range(len(self.layers_configuration) - 1):
            self.layers.append(torch.nn.Linear(self.layers_configuration[layer], self.layers_configuration[layer + 1]))
            self.layers.append(torch.nn.ReLU())
        self.layers.append(torch.nn.Linear(self.layers_configuration[-1], self.n_output))
        # self.dropout = nn.Dropout(0.001)

    def forward(self, x):
        out = self.layers(x)
        # out=self.dropout(out)
        return out

    def print_model(self):
        print(self.device)
        print(self.layers)

class TuningManagerMLP(MLP):
    def __init__(self, n_input, n_output, layers_configuration, name, train_loader, validation_loader,
                 existing_model=None):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        if self.device == "cuda":
            self.gpu_id = torch.cuda.current_device()

        print('INFO: USING DEVICE ' + self.device)

        self.name = name
        self.n_input = n_input
        self.n_output = n_output
        self.layers_configuration = layers_configuration

        super().__init__(n_input, n_output, layers_configuration)

        if existing_model is not None:
            self.load_state_dict(torch.load('./mlp/files/neural_networks/' + existing_model + '/model_' + existing_model + '.pt', weights_only=True))
            print("INFO: EXISTING MLP LOADED!")

        self.train_metrics = []
        self.validation_metrics = []
        self.optimizer = None
        self.loss_function = None
        self.nb_epochs = None
        self.batch_size = None
        self.is_set = False

        self.train_loader = train_loader
        self.validation_loader = validation_loader

    def save_mlp_parameters(self, neural_networks_path):
        save_path = neural_networks_path + self.name + '/'
        if not os.path.exists(save_path):
            os.mkdir(save_path)

        np.save(save_path + 'parameters.npy', np.array([self.n_input, self.n_output, self.layers_configuration],
                                                  dtype=object))

    def set_training_hyperparameters(self, nb_epochs, optimizer, learning_rate, loss_function):
        self.nb_epochs = nb_epochs

        # Define the loss function and optimizer

        if optimizer == 'adam':
            self.optimizer = torch.optim.Adam(self.parameters(), learning_rate)
        else:
            raise ValueError('optimizer parameter has to be specified')

        if loss_function == 'MSE':
            self.loss_function = torch.nn.MSELoss()
        else:
            raise ValueError('loss_function parameter has to be specified')

        self.is_set = True

    def save(self, neural_networks_path):
        save_path = neural_networks_path + self.name + '/'
        if not os.path.exists(save_path):
            os.mkdir(save_path)
        torch.save(self.state_dict(), save_path + 'model_' + self.name + '.pt')

        metrics = np.array([self.train_metrics, self.validation_metrics])
        np.save(save_path + 'metrics.npy', metrics)
        self.plot_metrics(save_path=save_path)

        print('INFO: EVERYTHING SAVED!')

    def plot_metrics(self, save_path=None):
        plt.figure()
        epoch = np.arange(self.nb_epochs) + 1
        plt.plot(epoch, self.train_metrics, label='Training', color='red')
        plt.plot(epoch, self.validation_metrics, label='Validation', color='green')
        plt.grid()
        plt.xlim([1, self.nb_epochs])
        plt.xlabel('Epoch')
        plt.ylabel('Squared $L2$ Norm')
        plt.legend()
        if save_path is not None:
            plt.savefig(save_path + 'metrics_plot.png')

    def start_training(self):
        # Run the training loop

        print('')
        print('INFO: STARTING TRAINING!')

        if not self.is_set:
            raise ValueError('Training Parameters have to be specified with set_training_parameters')

        for epoch in range(0, self.nb_epochs):
            print('')
            print('EPOCH ' + str(epoch + 1))

            self.layers.train()
            epoch_train_metrics = self._epoch_training()
            epoch_train_metrics = np.mean(epoch_train_metrics)
            print('Training Loss: ' + str(epoch_train_metrics))
            self.train_metrics.append(epoch_train_metrics)

            self.layers.eval()
            epoch_validation_metrics = self._epoch_validation()
            epoch_validation_metrics = np.mean(epoch_validation_metrics)
            print('Validation Loss: ' + str(epoch_validation_metrics))
            self.validation_metrics.append(epoch_validation_metrics)

        print('')
        print('INFO: TRAINING FINISHED!')

    def _epoch_training(self):
        print('TRAINING...')

        epoch_train_metrics = []

        for batch in tqdm.tqdm(self.train_loader, file=sys.stdout):
            inputs, targets = batch
            inputs, targets = inputs.float(), targets.float()

            self.optimizer.zero_grad()

            outputs = self.forward(inputs)

            loss = self.loss_function(outputs, targets)

            loss.backward()
            self.optimizer.step()

            epoch_train_metrics.append(float(loss.cpu().detach().numpy()))

        return np.array(epoch_train_metrics)

    def _epoch_validation(self):
        print('VALIDATION...')

        epoch_validation_metrics = []

        for test_batch in tqdm.tqdm(self.validation_loader, file=sys.stdout):
            inputs, targets = test_batch
            inputs, targets = inputs.float(), targets.float()

            outputs = self.forward(inputs)

            loss = self.loss_function(outputs, targets)

            epoch_validation_metrics.append(float(loss.cpu().detach().numpy()))

        return np.array(epoch_validation_metrics)

class MLPResults:
    def __init__(self, kinematics_obj, results_mlp):
        self.kinematics_obj = kinematics_obj

        self.lift_coefficient = results_mlp[:, 0]
        self.moment_coefficient = results_mlp[:, 1]

        self.rho = 1.225

        self.dynamic_pressure = (1 / 2) * self.rho * (self.kinematics_obj.freestream_velocity ** 2)
        self.lift = self.lift_coefficient * self.dynamic_pressure * self.kinematics_obj.chord
        self.moment = self.moment_coefficient * self.dynamic_pressure * (self.kinematics_obj.chord ** 2)

        self.heaving_power = self.lift * self.kinematics_obj.h_dot
        self.pitching_power = self.moment * self.kinematics_obj.theta_dot

        self.heaving_energy = results.compute_integral(x=self.kinematics_obj.time,
                                                       y=self.heaving_power)
        self.pitching_energy = results.compute_integral(x=self.kinematics_obj.time,
                                                        y=self.pitching_power)

        self.heaving_energy_recoverable = results.compute_integral(x=self.kinematics_obj.time,
                                                                   y=np.maximum(self.heaving_power, np.zeros(np.size(self.heaving_power, 0))))
        self.pitching_energy_recoverable = results.compute_integral(x=self.kinematics_obj.time,
                                                                    y=np.maximum(self.pitching_power, np.zeros(np.size(self.pitching_power, 0))))

        self.heaving_energy_consumed = results.compute_integral(x=self.kinematics_obj.time,
                                                                 y=np.minimum(self.heaving_power, np.zeros(np.size(self.heaving_power, 0))))
        self.pitching_energy_consumed = results.compute_integral(x=self.kinematics_obj.time,
                                                                 y=np.minimum(self.pitching_power, np.zeros(np.size(self.pitching_power, 0))))

        self.energy_recoverable = self.pitching_energy_recoverable + self.heaving_energy_recoverable
        self.energy_consumed = self.pitching_energy_consumed + self.heaving_energy_consumed

        self.energy_extracted = self.pitching_energy + self.heaving_energy
        # self.mean_power_extracted = np.mean(self.heaving_power[~np.isnan(self.heaving_power)] + self.pitching_power[~np.isnan(self.pitching_power)])
        self.mean_power_extracted = (1.0 / kinematics_obj.period) * results.compute_integral(x=self.kinematics_obj.time, y=(self.pitching_power + self.heaving_power))

        self.vertical_extent_airfoil_motion = 2.0 * self.kinematics_obj.heaving_amplitude
        self.flow_power = 0.5 * self.rho * (self.kinematics_obj.freestream_velocity ** 3) * self.vertical_extent_airfoil_motion * self.kinematics_obj.nb_period

        self.efficiency = self.mean_power_extracted / self.flow_power

def push_datasets_to_gpu(train_dataset, validation_dataset):
    train_dataset.X.cuda(torch.cuda.current_device())
    train_dataset.y.cuda(torch.cuda.current_device())
    validation_dataset.X.cuda(torch.cuda.current_device())
    validation_dataset.y.cuda(torch.cuda.current_device())


def load_parameters(neural_networks_path, model_name):
    parameters = np.load(neural_networks_path + model_name + '/parameters.npy', allow_pickle=True)
    n_input = parameters[0]
    n_output = parameters[1]
    layers_configuration = parameters[2]
    return n_input, n_output, layers_configuration


def validate(model, test_dataset, nb_time_steps_dataset, debug=False):
    r2_scores= []

    idx = np.arange(0, int(test_dataset.X.shape[0] / nb_time_steps_dataset)) * nb_time_steps_dataset

    for j in tqdm.tqdm(range(np.size(idx, 0))):
        y_prediction = [model(test_dataset.__getitem__(idx[j] + k)[0].float()).detach().numpy() for k in range(nb_time_steps_dataset)]
        y_truth = [test_dataset.__getitem__(idx[j] + k)[1].detach().numpy() for k in range(nb_time_steps_dataset)]

        y_prediction = np.array(y_prediction)
        y_truth = np.array(y_truth)

        r2_score = sklearn.metrics.r2_score(y_truth, y_prediction)

        if debug:
            plt.figure()
            plt.plot(y_prediction[:, 0], label='MLP | $C_L$')
            plt.plot(y_prediction[:, 1], label='MLP | $C_M$')
            plt.plot(y_truth[:, 0], label='Reference | $C_L$')
            plt.plot(y_truth[:, 1], label='Reference | $C_M$')
            plt.grid()
            plt.title('R2 Score: ' + str(r2_score))
            plt.legend()
            plt.show()

        if not np.all(y_truth == 0.0) and r2_score > 0.0:
            r2_scores.append(r2_score)

    r2_scores = np.array(r2_scores)
    mean_r2_score = np.mean(r2_scores)

    print("meanR2 = " + str(mean_r2_score))


