import multiprocessing
import os
import sys
from pymoo.core.problem import ElementwiseProblem
import numpy as np
from utils import kinematics, mlp, aeroelasticity, controller


class AeroMultiObjectivesProblem(ElementwiseProblem):
    def __init__(self, nn, xl, xu):
        super().__init__(n_var=3,
                         n_obj=2,
                         n_constr=1,
                         xl=xl,
                         xu=xu)
        self.nn = nn

    def _evaluate(self, x, out, *args, **kwargs):
        mlp_results = run_aerodynamical_model(self.nn, x)

        f1 = - mlp_results.energy_recoverable
        f2 = np.abs(mlp_results.energy_consumed)
        # f1 = - mlp_results.mean_power_extracted
        # f2 = - mlp_results.efficiency

        g1 = np.abs(mlp_results.energy_consumed) - np.abs(mlp_results.energy_recoverable)

        out["F"] = [f1, f2]
        out["G"] = [g1]


class AeroSingleObjectiveProblem(ElementwiseProblem):
    def __init__(self, nn, xl, xu, elementwise_runner):
        super().__init__(n_var=3,
                         n_obj=1,
                         n_constr=1,
                         xl=xl,
                         xu=xu,
                         elementwise_runner=elementwise_runner)
        self.nn = nn

    def _evaluate(self, x, out, *args, **kwargs):
        mlp_results = run_aerodynamical_model(self.nn, x)

        f1 = - mlp_results.efficiency

        g1 = np.abs(mlp_results.energy_consumed) - np.abs(mlp_results.energy_recoverable)

        out["F"] = [f1]
        out["G"] = [g1]


class DynMultiObjectivesProblem(ElementwiseProblem):
    def __init__(self, nn, xl, xu, elementwise_runner):
        super().__init__(n_var=3,
                         n_obj=2,
                         n_constr=1,
                         xl=xl,
                         xu=xu,
                         elementwise_runner=elementwise_runner)
        self.nn = nn

    def _evaluate(self, x, out, *args, **kwargs):
        efficiency, energy_recoverable, energy_consumed, power_recoverable, power_consumed = run_dynamical_model(self.nn, x)

        f1 = - (power_recoverable + power_consumed)
        f2 = - efficiency

        g1 = np.abs(energy_consumed) - np.abs(energy_recoverable)

        out["F"] = [f1, f2]
        out["G"] = [g1]


class DynSingleObjectiveProblem(ElementwiseProblem):
    def __init__(self, nn, xl, xu, elementwise_runner):
        super().__init__(n_var=3,
                         n_obj=1,
                         n_constr=0,
                         xl=xl,
                         xu=xu,
                         elementwise_runner=elementwise_runner)
        self.nn = nn

    def _evaluate(self, x, out, *args, **kwargs):
        efficiency, energy_recoverable, energy_consumed, power_recoverable, power_consumed = run_dynamical_model(self.nn, x)

        f1 = - efficiency

        out["F"] = [f1]
        # out["G"] = [g1]


def run_aerodynamical_model(model, variables):
    inputs_variables = ["x_f", "f", "theta_0"]
    outputs_variable = ["lift_coefficient", "moment_coefficient"]

    kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                           nb_time_step_per_period=1000,
                                           input_variable_names=inputs_variables,
                                           output_variable_names=outputs_variable,
                                           variable_values=[variables[2], variables[1], variables[0]],
                                           m=3,
                                           tau=9)

    mlp_outputs = model(kinematics_obj.get_inputs_mlp())

    mlp_outputs = mlp_outputs.detach().numpy()

    mlp_results = mlp.MLPResults(kinematics_obj, mlp_outputs)

    return mlp_results


def run_dynamical_model(model, variables):
    # Kinematics
    frequency = variables[0]

    pitching_amplitude = 79.5
    rotation_axis_in_percentage_chord = 0.32

    # Dynamical System
    m = 1.9
    K_h = variables[2]
    K_theta = 0.0
    C_h = variables[1]
    C_theta = 0.0

    inputs_variables = ["x_f", "f", "theta_0"]
    outputs_variable = ["lift_coefficient", "moment_coefficient"]

    kinematics_obj = kinematics.Kinematics(kinematics_reference=1,
                                           nb_time_step_per_period=1000,
                                           input_variable_names=inputs_variables,
                                           output_variable_names=outputs_variable,
                                           variable_values=[rotation_axis_in_percentage_chord, frequency,
                                                            pitching_amplitude],
                                           m=3,
                                           tau=9,
                                           nb_period=6)
    kinematics_obj.generate_kinematics()

    controller_obj = controller.PIDController(K_p=np.array([0.0, 500.0]),
                                              K_i=np.array([0.0, 75.0]),
                                              K_d=np.array([0.0, 10.0]),
                                              mode='semi_active')

    dynamical_system_obj = aeroelasticity.DynamicalSystem(kinematics_obj=kinematics_obj,
                                                          model=model,
                                                          controller=controller_obj,
                                                          m=m,
                                                          K_h=K_h,
                                                          K_theta=K_theta,
                                                          C_h=C_h,
                                                          C_theta=C_theta)

    try:
        dynamical_system_obj.motion_solver()

        efficiency = dynamical_system_obj.efficiency
        energy_recoverable = dynamical_system_obj.energy_recoverable
        energy_consumed = dynamical_system_obj.energy_consumed

        power_recoverable = dynamical_system_obj.power_recoverable
        power_consumed = dynamical_system_obj.power_consumed

    except:
        efficiency = 0.0
        energy_recoverable = 0.0
        energy_consumed = 0.0

        power_recoverable = 0.0
        power_consumed = 0.0

    if 0.0 <= efficiency <= 1.0:
        pass
    else:
        efficiency = 0.0

    return efficiency, energy_recoverable, energy_consumed, power_recoverable, power_consumed


def save_results(design_space, optimization_functions, bounds, path, model_name, optimization_name):
    save_path = path + model_name + '/'
    save_optimization_path = save_path + optimization_name + '/'

    optimization_results = np.array([design_space, optimization_functions, bounds], dtype=object)

    if not os.path.exists(save_path):
        os.mkdir(save_path)

    if not os.path.exists(save_optimization_path):
        os.mkdir(save_optimization_path)

    np.save(save_optimization_path + 'optimization_results.npy', optimization_results)


def load_results(path, model_name, optimization_name):
    load_path = path + model_name + '/'
    load_optimization_path = load_path + optimization_name + '/'
    optimization_results = np.load(load_optimization_path + 'optimization_results.npy', allow_pickle=True)

    design_space = optimization_results[0]
    optimization_functions = optimization_results[1]
    bounds = optimization_results[2]

    return design_space, optimization_functions, bounds


