import numpy as np
import torch

def state_space_func(t, x, kinematics_obj, model, controller, time, m, b, C_h, C_theta, K_h, K_theta, rho):
    t = np.round(t, 12)

    c = kinematics_obj.chord
    x_f_c = kinematics_obj.rotation_axis_in_percentage_chord
    x_f = x_f_c * c

    surface = kinematics_obj.chord * b

    I_theta = (1.0 / 3.0) * m * (c ** 2 - (3.0 * c * x_f) + 3.0 * (x_f ** 2))
    S_y_theta = m * ((c / 2.0) - x_f)

    if t == 0.0:
        state_space_func.time = np.array([])
        state_space_func.h = np.array([])
        state_space_func.theta = np.array([])
        state_space_func.h_dot = np.array([])
        state_space_func.theta_dot = np.array([])
        state_space_func.lift = np.array([])
        state_space_func.moment = np.array([])
        state_space_func.control_lift = np.array([])
        state_space_func.control_moment = np.array([])
        state_space_func.lift = np.array([])
        state_space_func.moment = np.array([])
        state_space_func.stiffness_h_force = np.array([])
        state_space_func.stiffness_theta_force = np.array([])
        state_space_func.damping_h_force = np.array([])
        state_space_func.damping_theta_force = np.array([])

    M = np.array([[1.0, 0.0, 0.0, 0.0],
                  [0.0, m, 0.0, S_y_theta],
                  [0.0, 0.0, 1.0, 0.0],
                  [0.0, S_y_theta, 0.0, I_theta]])

    K = np.array([[0.0, 0.0, 0.0, 0.0],
                  [- K_h, 0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 0.0],
                  [0.0, 0.0, - K_theta, 0.0]])

    C = np.array([[0.0, 0.0, 0.0, 0.0],
                  [0.0, - C_h, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0, - C_theta]])

    I = np.array([[0.0, 1.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 1.0],
                  [0.0, 0.0, 0.0, 0.0]])

    if np.isin(np.round(t, 8), np.round(time, 8)) and not np.isin(np.round(t, 8), np.round(state_space_func.time, 8)):
        state_space_func.time = np.append(state_space_func.time, t)
        state_space_func.h = np.append(state_space_func.h, x[0])
        state_space_func.theta = np.append(state_space_func.theta, x[2])
        state_space_func.h_dot = np.append(state_space_func.h_dot, x[1])
        state_space_func.theta_dot = np.append(state_space_func.theta_dot, x[3])

        state_space_func.stiffness_h_force = np.append(state_space_func.stiffness_h_force, np.dot(K, x)[1])
        state_space_func.stiffness_theta_force = np.append(state_space_func.stiffness_theta_force, np.dot(K, x)[3])
        state_space_func.damping_h_force = np.append(state_space_func.damping_h_force, np.dot(C, x)[1])
        state_space_func.damping_theta_force = np.append(state_space_func.damping_theta_force, np.dot(C, x)[3])

        reference = [kinematics_obj.kinematics_function(t)[1], kinematics_obj.kinematics_function(t)[0]]
        control_forces_output = controller.compute(state_variable=x,
                                                   reference=reference,
                                                   dt=kinematics_obj.dt)

        state_space_func.control_lift = np.append(state_space_func.control_lift, control_forces_output[0])
        state_space_func.control_moment = np.append(state_space_func.control_moment, control_forces_output[1])

        if t < kinematics_obj.dt * (kinematics_obj.m - 1) * kinematics_obj.tau:
            state_space_func.lift = np.append(state_space_func.lift, 0.0)
            state_space_func.moment = np.append(state_space_func.moment, 0.0)

        else:
            inputs = torch.tensor([kinematics_obj.rotation_axis_in_percentage_chord,
                                   kinematics_obj.frequency,
                                   x[2],
                                   x[3],
                                   x[0],
                                   x[1],
                                   state_space_func.theta[-9],
                                   state_space_func.theta_dot[-9],
                                   state_space_func.h[-9],
                                   state_space_func.h_dot[-9],
                                   state_space_func.theta[-18],
                                   state_space_func.theta_dot[-18],
                                   state_space_func.h[-18],
                                   state_space_func.h_dot[-18]],
                                  dtype=torch.float32)

            aerodynamic_coefficients = model(inputs).detach().numpy()

            lift = 0.5 * rho * (kinematics_obj.freestream_velocity ** 2) * surface * aerodynamic_coefficients[0]

            moment = 0.5 * rho * (kinematics_obj.freestream_velocity ** 2) * surface * kinematics_obj.chord * aerodynamic_coefficients[1]

            state_space_func.lift = np.append(state_space_func.lift, lift)
            state_space_func.moment = np.append(state_space_func.moment, moment)

    control_forces = np.array([0.0, state_space_func.control_lift[-1], 0.0, state_space_func.control_moment[-1]])
    # else:
    #     control_forces = np.array([0.0, control_forces_output[0], 0.0, control_forces_output[1]])

    aerodynamic_forces = np.array([0.0, state_space_func.lift[-1], 0.0, state_space_func.moment[-1]])

    x_dot = np.dot(np.linalg.inv(M), (np.dot(K, x) + np.dot(C, x) + np.dot(I, x) + control_forces + aerodynamic_forces))

    return x_dot